# sbmate.py
# calculate annotation scores

import pandas as pd
from SBMate import sbml_annotation as sa
from SBMate.metric_calculator import MetricCalculator


class AnnotationMetrics(object):
  """
  Collects model annotations
  calculates three scores:
  1. coverage
  2. consistency
  3. specificity

  Attributes
  ----------
  annotations: sbml_annotation.SortedSBMLAnnotation
      Sorted annotations for each knowledge resource.
  calculatorDf: dataframe of metrics.
  """

  def __init__(self, model_file=None, metric_calculator_classes=None):
    """
    Parameters
    ----------
    model_file: str
        Address/name of the .xml model file
    metric_calculator_classes: list-type
    """
    # model file can take None
    if model_file is None:
      self.annotations = None
      self.metrics_df = None
    else:
      self.annotations = sa.SortedSBMLAnnotation(file=model_file)
      if metric_calculator_classes is None:
        metric_calculator_classes = []
      metric_calculator_classes.append(MetricCalculator)
      # Calculate a DataFrame for each metric calculator
      dfs = []
      # in case model_file is given as a path, get the last file name
      index_model_name = model_file.split('/')[-1]
      for cls in metric_calculator_classes:
        calculator = cls(annotations=self.annotations, model_name=index_model_name)
        dfs.append(calculator.calculate())
      # Merge the DataFrames
      self.metrics_df = pd.concat(dfs, axis=1)

  def _getMetricsReport(self):
    """
    Create a string report for
    an AnnotationMetrics class.

    Returns
    -------
    '': str
        Report summarizing the metrics df.
    """
    report = ["Summary of Metrics (%s)\n----------------------\n"
        % self.metrics_df.index[0]]
    report = report + ["%s: %s\n" % (col, self.metrics_df[col][0])
        for col in self.metrics_df]
    report.append("----------------------\n")
    return ('').join(report)

  @classmethod
  def getMetrics(cls, file, output="report"):
    """
    Using the AnnotationMetrics class,
    produces report on the three metrics.

    Parameters
    ----------
    file: str/str-list
        Address(es) of model file (.xml).
        Should be string or list of string.
    output: str
        The type of output ("report" or "table").

    Returns
    --------
    res: str / pandas.DataFrame / None
        Final report (summary) of the model.
        Return None if input type is incorrect.
    """

    # Throws ValueError if the file argument type is incorrect. 
    flag = False
    if isinstance(file, str):
      file_list = [file]
    elif isinstance(file, list):
      if all([isinstance(one_file, str) for one_file in file]):
        file_list = file
      else:
        flag = True
    else:
      flag = True
    if flag:
      raise ValueError("Should be a valid file name.")

    annotation_metrics_list = [cls(model_file=one_file) 
                               for one_file in file_list]
    if output == "report":
      res_list = [m._getMetricsReport() for m in annotation_metrics_list]
      res = ('\n').join(res_list)
    elif output=="table":
      res_list = [m.metrics_df for m in annotation_metrics_list]
      res = pd.concat(res_list)
    return res



# uniprot_kegg_analyzer.py
"""
Analyzer class for non-DAG knowledge resources
such as UNIPROT and KEGG.
The two classes calculate consistency & specificity, 
and returns appropriate values. 
"""

import collections
import libsbml
import numpy as np
import os
import re
import requests
from SBMate import constants as cn


ONT_TO_URL= {"uniprot":"https://www.uniprot.org/uniprot/",
             "kegg_process":"https://www.genome.jp/entry/",
             "kegg_species":"https://www.genome.jp/entry/",
            }
KEGG_ERROR_MESSAGE = "No such data was found"


class NonDAGAnalyzer(object):
  """
  Analyzer for UNIPROT and KEGG.

  Attributes
  ----------
  term_id: str-list
      List of identifiers to analyze.
  ontology: str
      Appropriate ontology type.
  object_type: libsbml.AutoProperty
      Type of the model entity.
  consistent: bool
      Bool determining whether the entity-object is consistent.     

  Methods
  -------
  getOneTermConsistency (one_term)
      Check if the given term (ontology identifier)
      is consistent, using url. 
  getConsistency (inp_term)
      Check if the given term (ontology identifier)
      is consistent.      
  getSpecificity (inp_term)
      Get specificity of a list of terms.
      These ontologies are 1.0 if consistent. 
  """

  def __init__(self, term_id, ontology,
               object_type):
    """
    Parameters
    ----------
    term_id: str-list
        Identifier to analyze
    ontology: str
        Ontology (knowledge resource).
        Should be one of {'go', 'sbo', 'chebi'}.
    object_type: libsbml.AutoProperty
        Type of model entity. For example, libsbml.Reactionß
    """
    self.term_id = term_id
    self.ontology = ontology
    self.object_type = object_type
    self.consistent = self.getConsistency(inp_term=self.term_id)


  def getOneTermConsistency(self, one_term):
    """
    Get consistency of one term,
    by connecting to the url
    and checking it works. 

    Parameters
    ----------
    one_term: str
        One identifier string to check.

    Returns
    -------
    '': bool
        True if consistent; otherwise False
    """
    r = requests.get(ONT_TO_URL[self.ontology]+one_term)
    # for kegg, needs to check whether the text below is in the page
    if KEGG_ERROR_MESSAGE in r.text:
      return False
    else:
      return r.ok

  def getConsistency(self, inp_term):
    """
    Using getOneTermConsistency(),
    check if the terms are consistent.

    Parameters
    ----------
    inp_term: str/str-list
        A list of terms,
        in the same ontology,
        from a single model entity. 

    Returns
    -------
    '': bool
        True if all terms are consistent; otherwise False.
    """
    # First, check if the input term is in correct format
    if isinstance(inp_term, str):
      inp_list = [inp_term]
    elif isinstance(inp_term, list):
      # if it is a list, check whether all terms are string
      if sum([isinstance(t, str) for t in inp_term])==len(inp_term):
        inp_list = inp_term
      else:
        return False
    else:
      return False
    # second, check if the object type and ontology match
    if not self.ontology in cn.OBJECT_ONT_MAP_FILT[self.object_type]:
      return False
    # finally, query the term and get consistency
    res = [self.getOneTermConsistency(t) for t in inp_list]
    if all(res):
      return True
    else:
      return False

  def getSpecificity(self, inp_term):
    """
    Get specificity score of one or more terms.
    Specificity is 1 for uniprot and KEGG, 
    as long as the therms are consistent.

    Parameters
    ----------
    inp_term: str/str-list
        List of identifiers to check specificity

    Returns
    -------
    '': float/None
        If consistent, 1.0; otherwise None.
    """
    if not self.consistent:
      return None
    else:
      return float(1.0)





# metric_calculator.py
# calculate annotation scores

import libsbml
import networkx as nx
import numpy as np
import os
import pandas as pd
import re
import requests
from SBMate import constants as cn
from SBMate import dag_analyzer as da
from SBMate import uniprot_kegg_analyzer as uka

# mapping reaction type to appropriate analyzer class
ANALYZER_DICT = {'go': da.DAGAnalyzer,
                 'sbo': da.DAGAnalyzer,
                 'chebi': da.DAGAnalyzer,
                 'kegg_species': uka.NonDAGAnalyzer,
                 'kegg_process': uka.NonDAGAnalyzer,
                 'uniprot': uka.NonDAGAnalyzer,
                }


class MetricCalculator(object):
  """
  calculates three scores:
  1. coverage
  2. consistency
  3. specificity

  Attributes
  ----------
  annotations: sbml_annotation.SortedSBMLAnnotation 
      Sorted annotations for each knowledge resource.
  annotated_entities: str-list
      List of model entity names that are annotated.
  consistent_entities: dict
      Dictionary of model entity name: list of consistent analyzer classes.
  coverage: float
      Model coverage score.
  consistency: float
      Model consistency score. 
  specificity: float
      Model specificity score. 

  Methods
  -------
  getCoverage(annotated_entities)
      Calculates model coverage score.
  getConsistency
      Calculates model consistency score.
  getSpecificity
      Calculates model specificity score. 
  """

  def __init__(self, annotations, model_name):
    """
    Parameters
    ----------
    annotations: sbml_annotation.SortedSBMLAnnotation 
        Sorted annotations for the five knowledge resources.
    model: str
        Name of the model; will be index of the dataframe
    """
    self.annotations = annotations
    self.model_name = model_name

  def calculate(self):
    """
    Creates the metrics.

    Returns
    -------
    '': pd.DataFrame
        Merged dataframe for each calculated metric.
    """
    if self.annotations.annotations:
      len_annotatable_entities = len(self.annotations.annotations)
    else:
      len_annotatable_entities = 0.0
    annotated_entities, coverage = self._getCoverage()
    if annotated_entities:
      len_annotated_entities = len(annotated_entities)
    else:
      len_annotated_entities = None
    # calculate scores using the two types of Analyzer class
    consistent_entities, consistency = self._getConsistency(annotated_entities)
    if consistent_entities:
      len_consistent_entities = len(consistent_entities)
    else:
      len_consistent_entities = None
    specificity = self._getSpecificity(consistent_entities)
    entities_Df = self.mkDataframe({'annotatable_elements': len_annotatable_entities})
    coverage_Df = self.mkDataframe({'annotated_elements':len_annotated_entities,
                                    'coverage': coverage})
    consistency_Df = self.mkDataframe({'consistent_elements':len_consistent_entities,
                                       'consistency': consistency})
    specificity_Df = self.mkDataframe({'specificity': specificity})

    return pd.concat([entities_Df, coverage_Df, consistency_Df, specificity_Df], axis=1)

  def mkDataframe(self, score_info):
    """
    Creates a DataFrame using the entity information
    and the scores.

    Parameters
    ----------
    score_info: dict
        Dictionary of the score, number of entities, etc. 

    Returns
    -------
    '': pandas.DataFrame
        Dataframe including entities and the score. 
    """
    return pd.DataFrame(score_info, index=[self.model_name])


  def _getConsistency(self, annotated_entities):
    """
    Calculate consistency using analylzer classes. 
    Consistency is defined as:
    (# consistently annotated entities) / (# annotated entities)
    'annotated_entities' is a list of annotated model entity names.
    Returns dictionary of entity & analyzer instances,
    as well as consistency score.

    Parameters
    ----------
    annotated_entities: str-list 
        List of annotated entities.

    Returns
    -------
    consistent_dicts: dict (str: Analyzer-list) / None
        Dictionary of consistenty entities and list of analyzers.
    consistency_score: float/None
        Consistency score. None if no object is annotated.
    """
    if not annotated_entities:
      return None, None
    consistent_entities = []
    consistent_dicts = dict()
    for anot_key in annotated_entities:
      one_anot = self.annotations.annotations[anot_key]
      # annotated entities
      one_entity_onts = [ANALYZER_DICT[key](term_id=one_anot[key],
                                            ontology=key,
                                            object_type=one_anot['object_type']) \
                         for key in ANALYZER_DICT if one_anot[key]]

      # if consistent for all ontologies, update list of consistent objects
      if all([r.consistent for r in one_entity_onts]):
        consistent_dicts[anot_key] = one_entity_onts
    consistency_score = np.round(len(consistent_dicts.keys())/len(annotated_entities), 2)
    return consistent_dicts, consistency_score


  def _getSpecificity(self, consistent_entities):
    """
    Calculate specificity using analyzer classes.
    Specificity is defined as:
    log(num_ancestors/num_all_nodes) / log(1/num_all_nodes)
    'consistent_entities' is a list of consistent model entities. 

    Parameters
    ----------
    consistent_entities: dict (str: Analyzer-list)
         Dictionary of consistenty entities and list of analyzers.

    Returns
    -------
    specificity_score: float
        Model specificity score.
    """
    # calculates specificity only if it has at least one consistent entity
    if consistent_entities:
      entities_specificity = [np.mean([one_analyzer.getSpecificity(one_analyzer.term_id) for \
                                       one_analyzer in consistent_entities[one_key]]) for \
                              one_key in consistent_entities.keys()]
      specificity_score = np.round(np.mean(entities_specificity), 2)
      return specificity_score
    else:
      return None

  def _getCoverage(self):
    """
    Coverage is defined as:
    (# annotated entities) / (# annotatable entities).

    Returns
    -------
    list_annotated_entities: str-list
        List of annotated model entity names.
    '': float
        Coverage score. 
    """
    num_annotatable_entities = len(self.annotations.annotations)
    list_annotated_entities = [k for \
                               k in \
                               self.annotations.annotations.keys() \
                               if any([self.annotations.annotations[k][ont] for ont \
                               in cn.KNOWLEDGE_TYPES_REP])]
    num_annotated_entities = len(list_annotated_entities)
    coverage_score = float(num_annotated_entities/num_annotatable_entities)
    return list_annotated_entities, np.round(coverage_score, 2)



# sbml_annotation.py

import collections
import libsbml
import os
import re
from SBMate import constants as cn

ObjectAnnotation = collections.namedtuple('ObjectAnnotation',
                                        ['id', 'object_type', 'annotation'],
                                        )


class RawSBMLAnnotation(object):
  """
  Collection of SBML object Annotations,
  from sbo_term and getAnnotationString() methods.

  Attributes
  ----------
  sbo: namedtuple 'ObjectAnnotation' - list
      Collection of SBO terms
  str_annotation: namedtuple 'ObjectAnnotation' - list
      Collection of string annotations,
      bqbiol:is or bqbiol:isVersionOf.

  Methods
  -------
  getSBOAnnotation (sbml_object)
      Create a list of SBO terms
      from .sbo_term attribute of
      libsbml model entity.
  getOntAnnotatoin (sbml_object)
      Create a list of string annotations
      from .getAnnotataionString(). method of
      libsbml model entity. 
  """

  def __init__(self, input_file,
               select_objects=cn.BIOMODEL_OBJECTS):
    """
    Parameters
    ----------
    input_file: str
        Name/location of model file (.xml)
    select_objects: libsbml.AutoProperty - list
        List of objects to pull out annotations from.
    """
    # load sbml file
    reader = libsbml.SBMLReader()
    document = reader.readSBML(input_file)
    # only keep the objects of interest
    model_objects = [ele for ele in document.getListOfAllElements() \
                     if isinstance(ele, tuple(select_objects))] 
    self.sbo = [self.getSBOAnnotation(ele) for ele in model_objects]
    self.str_annotation = [self.getOntAnnotation(ele) for ele in model_objects]

  def getSBOAnnotation(self, sbml_object):
    """
    Returns a proper SBO term (e.g., SBO:0000290) 
    using the given sbml object.
    Return None if term is not given.

    Parameters
    ----------
    sbml_object: libsbml.AutoProperty

    Returns
    -------
    '': namedtuple 'ObjectAnnotation': (id, type, str/None)
    """
    def formatSBO(sbo_num):
      """
      Reformat an SBO term into str.
      Return None if -1 (not provided).

      Parameters
      ----------
      sbo_num: int

      Returns
      -------
      '': str/None
          Return None if sbo term is -1 
          (i.e.,  not provided)
      """
      if sbo_num == -1:
        return None
      else:
        return 'SBO:' + format(sbo_num, '07d')
    #
    input_id = sbml_object.getId()
    input_type = type(sbml_object)
    input_sbo = formatSBO(sbml_object.sbo_term)
    #
    return ObjectAnnotation(input_id, input_type, input_sbo)
    
  def getOntAnnotation(self, sbml_object):
    """
    Parse string and return string annotation,
    marked as <bqbiol:is> or <bqbiol:isVersionOf>.
    If neither exists, return None/

    Parameters
    ----------
    sbml_object: libsbml.AutoProperty

    Returns
    -------
    '': namedtuple 'ObjectAnnotation': (id, type, str/None)
    """
    input_id = sbml_object.getId()
    input_type = type(sbml_object)
    input_annotation = sbml_object.getAnnotationString()
    #
    is_str = ''
    isVersionOf_str = ''
    is_str_match = re.findall('<bqbiol:is[^a-zA-Z].*?<\/bqbiol:is>',
                              input_annotation,
                              flags=re.DOTALL)
    if len(is_str_match)>0:
      is_str_match_filt = [s.replace("      ", "") for s in is_str_match]
      is_str = '\n'.join(is_str_match_filt)

    is_VersionOf_str_match = re.findall('<bqbiol:isVersionOf[^a-zA-Z].*?<\/bqbiol:isVersionOf>',
                                        input_annotation,
                                        flags=re.DOTALL)
    #
    if len(is_VersionOf_str_match) > 0:
      is_VersionOf_str_match_filt = [s.replace("      ", "") for s in is_VersionOf_str_match]
      isVersionOf_str = '\n'.join(is_VersionOf_str_match_filt)
    #
    combined_str = is_str + isVersionOf_str
    if combined_str == '':
      combined_str = None
    return ObjectAnnotation(input_id, input_type, combined_str)



class SortedSBMLAnnotation(object):
  """
  SBML annotations sorted by type.
  SBO, GO, KEGG, CHEBI, UNIPROT. 

  Attributes
  ----------
  raw_annotations: RawSBMLAnnotation
      Instance of RawSBMLAnnotation class.
  object_ids: str-list
      Names of model entities. 
  annotations: dict(dict)
      Dictionary of dictionary. 
      object_id: {ontology type: identifiers}

  Methods
  -------
  getAnnotoationDict (input_id)
      Get dictionary of annotations.
  getKnowledgeResourceTuple (input_annotation)
      Extract identifiers from URI, 
      included in the string annotation. 
  """

  def __init__(self, file, knowledge_resources=cn.KNOWLEDGE_TYPES_REP):
    # For now, use default biomodel objects.
    self.raw_annotations = RawSBMLAnnotation(input_file=file)
    self.object_ids = [ele.id for ele in self.raw_annotations.str_annotation]
    self.annotations = {one_id:self.getAnnotationDict(one_id) for one_id in self.object_ids}

  def getAnnotationDict(self, input_id):
    """
    Get dictionary of annotations for an object,
    where the key is object id
    and the items are each annotation
    per category. 
    Returns a nested dictionary
    for each object. 

    Parameters
    ----------
    input_id: str
        Model entity id (name) to get annotation of.

    Returns
    -------
    annotation_dict: dict {dict: {resource:identifier}}
        Nested dictionary of annotations per object. 
    """
    annotation_dict = dict.fromkeys(cn.KNOWLEDGE_TYPES_REP)
    str_annotation_item = [ele for ele in self.raw_annotations.str_annotation \
                           if ele.id==input_id][0]
    str_annotation_tuples = self.getKnowledgeResourceTuple(str_annotation_item.annotation)
    if str_annotation_tuples:
      tup_keys = list(set([cn.KNOWLEDGE_TYPES_DCT[ele[0]] for ele in str_annotation_tuples]))
      for one_key in tup_keys:
        vals = [ele[1] for ele in str_annotation_tuples if cn.KNOWLEDGE_TYPES_DCT[ele[0]]==one_key]
        annotation_dict[one_key] = vals
    
    # extra formatting for sbo
    def getSBOForm(inp_sbo):
      """
      Reformat an SBO term into str.
      Return None if -1 (not provided).

      Parameters
      ----------
      inp_sbo: int

      Returns
      -------
      res_sbo: str/None
          Return None if sbo term is -1 
          (i.e.,  not provided)
      """
      m = re.search('[0-9]+', inp_sbo)
      if m:
        res_sbo = 'SBO:' + inp_sbo[m.start():m.end()]
        return res_sbo
      else:
        return None
      
    if annotation_dict['sbo']:
      annotation_dict['sbo'] = [getSBOForm(ele) for ele in annotation_dict['sbo']]

    # get sbo term from .sbo_term
    sbo_item = [ele for ele in self.raw_annotations.sbo \
                if ele.id==input_id][0]
    if sbo_item.annotation:
      # check if sbo term is also given as string
      if annotation_dict['sbo']:
        annotation_dict['sbo'] = annotation_dict['sbo'].append(sbo_item.annotation)
      else:
        annotation_dict['sbo'] = [sbo_item.annotation]
    # finally, add the object id and type
    annotation_dict['object_id'] = input_id
    annotation_dict['object_type'] = str_annotation_item.object_type
    
    return annotation_dict

  def getKnowledgeResourceTuple(self, input_annotation):
    """
    Extract all annotation type tuple from URIs 
    marked with identifier.org.
    If nothing exists, return None

    Parameters
    ----------
    input_annotation: str
        Annotation string to extract annotation URI from.

    Returns
    -------
    '': str-tuple/None
        Extracted annotation.
        Ontology - identifier tuple.
    """
    if input_annotation:
      identifiers_list = re.findall('identifiers\.org/.*/', input_annotation)
      return [(r.split('/')[1],r.split('/')[2].replace('\"', '')) \
              for r in identifiers_list \
              if r.split('/')[1] in cn.ALL_KNOWLEDGE_TYPES]
    else:
      return None










# constants.py
"""
Constants for modlues
"""

import libsbml
import os

PROJECT_NAME = "SBMate"

# folders
PROJECT_DIR = os.path.dirname(os.path.abspath(__file__))
# one level higher
PROJECT_DIR = os.path.dirname(PROJECT_DIR)
RESOURCE_DIR = os.path.join(PROJECT_DIR, "SBMate/knowledge_resources")
TEST_DIR = os.path.join(PROJECT_DIR, "tests")


# Model entity types
MODEL = libsbml.Model
REACTION = libsbml.Reaction
SPECIES = libsbml.Species
COMPARTMENT = libsbml.Compartment
BIOL_PROC = 'biological_process'
MOLE_FUNC = 'molecular_function'
CELL_COMP = 'cellular_component'
# GO term ids matching the above names
GO_BIO_PROC = "GO:0008150"
GO_MOL_FUNC = "GO:0003674"
GO_CELL_COMP = "GO:0005575"
# GO_PARENTS = [GO_BIO_PROC, GO_MOL_FUNC, GO_CELL_COMP]
GO_ROOTS = [GO_BIO_PROC, GO_MOL_FUNC, GO_CELL_COMP]

BIOMODEL_OBJECTS = [MODEL,
                    REACTION,
                    SPECIES,
                    COMPARTMENT,
                   ]
OBJECT_GO_MAP = {MODEL: {BIOL_PROC, MOLE_FUNC},
                 REACTION: {BIOL_PROC, MOLE_FUNC},
                 SPECIES: {CELL_COMP},
                 COMPARTMENT: {CELL_COMP},
                }
GO_ROOT_DICT = {GO_BIO_PROC: BIOL_PROC,
                GO_MOL_FUNC: MOLE_FUNC,
                GO_CELL_COMP: CELL_COMP,
                None: "None"
                }

# Knowledge resource identifiers for URI
GOS = {'go', 'GO', 'obo.go'}
SBOS = {'biomodels.sbo', 'sbo'}
CHEBIS = {'chebi', 'obo.chebi'}
KEGG_SPEC = {'kegg.compound',
             'kegg.genes',
             'kegg.drug',
             'kegg.orthology'}
KEGG_PROC= {'kegg.reaction',
            'kegg.pathway'}
UNIPROTS = {'uniprot', 'uniprot.isoform'}

ALL_KNOWLEDGE_TYPES = set.union(*[GOS, SBOS, CHEBIS,
                                  KEGG_SPEC, KEGG_PROC,
                                  UNIPROTS])
KNOWLEDGE_TYPES_INT = [GOS, SBOS,
                       CHEBIS, KEGG_SPEC,
                       KEGG_PROC, UNIPROTS]
KNOWLEDGE_TYPES_REP = ['go', 'sbo',
                       'chebi', 'kegg_species',
                       'kegg_process', 'uniprot']
KNOWLEDGE_TYPES_ZIP = zip(KNOWLEDGE_TYPES_INT, KNOWLEDGE_TYPES_REP)
KNOWLEDGE_TYPES_DCT = dict()
for ele in KNOWLEDGE_TYPES_ZIP:
  KNOWLEDGE_TYPES_DCT.update(dict.fromkeys(ele[0], ele[1]))

# do we need this at this point? 
ALL_KNOWLEDGE_TERMS = {'go':GOS,
					   'sbo': SBOS,
					   'chebi': CHEBIS,
					   'kegg_species': KEGG_SPEC,
					   'kegg_process': KEGG_PROC,
					   'uniprot': UNIPROTS,
					   }
# items are a set - union of multiple sets
# OBJECT_ONT_MAP_RAW = {MODEL: set.union(*[GOS, SBOS, KEGG_PROC]),
#                       REACTION: set.union(*[GOS, SBOS, KEGG_PROC]),
#                       SPECIES: set.union(*[KEGG_SPEC, UNIPROTS]),
#                       COMPARTMENT: set.union(*[KEGG_SPEC, UNIPROTS]),
#                      }

# physical entity.. and occuring entity.. is from SBO

PROCESS_ROOTS = {'biological_process', 'molecular_function',
                 'kegg_process', 'occurring entity representation'}
# COMPONENT_TERMS = {'cellular_component', 'kegg_species', 'chemical entity',
#                    'uniprot', 'physical entity representation'}
SPECIES_ROOTS = {'cellular_component', 'kegg_species', 'chemical entity',
                   'uniprot', 'physical entity representation'}
COMPARTMENT_ROOTS = {'cellular_component', 'physical entity representation'}            
# when konwledge resources were mapped to KNOWLEDGE_TYPES_REP
OBJECT_ONT_MAP_FILT = {MODEL: PROCESS_ROOTS,
                       REACTION: PROCESS_ROOTS,
                       SPECIES: SPECIES_ROOTS,
                       COMPARTMENT: COMPARTMENT_ROOTS,
                      }


# SBO terms
# categorical parents. Children of 'SBO:0000000'. 
MATH_EXP = 'SBO:0000064'
META_REP = 'SBO:0000544'
MODEL_FRAME = 'SBO:0000004'
ENTITY_REP = 'SBO:0000231'
PART_ROLE = 'SBO:0000003'
PHYSICAL_ENT = 'SBO:0000236'
SYS_PARAM = 'SBO:0000545'

SBO_ROOTS = [MATH_EXP, META_REP,
             MODEL_FRAME, ENTITY_REP,
             PART_ROLE, PHYSICAL_ENT,
             SYS_PARAM]

SBO_ROOT_DICT = {MATH_EXP:'mathematical expression',
                META_REP:'metadata representation',
                MODEL_FRAME:'modelling framework',
                ENTITY_REP:'occurring entity representation',
                PART_ROLE:'participant role',
                PHYSICAL_ENT:'physical entity representation',
                SYS_PARAM:'systems description parameter',
                None: "None"
                }



# CHEBI terms
CHEMICAL_ENTITY = "CHEBI:24431"
CHEBI_ROOTS = [CHEMICAL_ENTITY]
CHEBI_ROOT_DICT = {CHEMICAL_ENTITY: "chemical entity",
                   None: "None"
                   }

# DAG ontologies (GO, SBO, CHEBI)
DAG_ROOT_MAP = {"go":GO_ROOT_DICT,
                "sbo":SBO_ROOT_DICT, 
                "chebi":CHEBI_ROOT_DICT}



























# dag_analyzer.py
"""
Analyzer class for ontology terms.
This class is for DAG-type knowledge resources, 
i.e.,  GO, SBO, and CHEBI.
The two classes calculate consistency & specificity, 
and returns appropriate values. 
"""

import collections
import libsbml
import networkx as nx
import numpy as np
import os
import pickle5 as pickle
from SBMate import constants as cn

# Load ontology graphs
SBO_G = pickle.load(open(os.path.join(cn.RESOURCE_DIR, "sbo_graph.gpickle"), 'rb'))
CHEBI_G = pickle.load(open(os.path.join(cn.RESOURCE_DIR, "chebi_graph.gpickle"), 'rb'))
GO_G = pickle.load(open(os.path.join(cn.RESOURCE_DIR, "go_graph.gpickle"), 'rb'))

# Mapping ontology into specific graph
ONT_TO_G = dict({"go":GO_G, "sbo":SBO_G, "chebi":CHEBI_G})
# Mapping ontology to list of roots
ONT_TO_ROOT = dict({"go": cn.GO_ROOTS, "sbo":cn.SBO_ROOTS, "chebi":cn.CHEBI_ROOTS})


class DAGAnalyzer(object):
  """
  Analyzer for GO, SBO, and CHEBI.
  Calculates object-level consistency 
  and specificity. 
  Terms are hierarchically arranged,
  having one or a few terms as roots for
  the rest of the terms. 

  Attributes
  ----------
  term_id: str-list
      List of identifiers to analyze.
  ontology: str
      Appropriate ontology type.
  object_type: libsbml.AutoProperty
      Type of the model entity.
  dag: networkx.DiGraph
      Directed Acyclic Graph for ontology system.
  possible_roots: str-list
      List of identifiers (roots) for all terms 
      within the system.
  term_to_root: None -> dict
      Dictionary, mapping each identifier: its root.
  consistent: bool
      Bool determining whether the entity-object is consistent.   

  Methods
  -------
  findRoot(inp_term)
      Find an appropriate root of one identifier.
  getConsistency (inp_term)
      Check if the given term (ontology identifier)
      is consistent.
  getOneTermSpecificity (one_term)
      Get the specificity of one identifier.
  getSpecificity (inp_term)
      Calculate specificity of a list of terms,
      using getOneTermSpecificity method.
  """

  def __init__(self, term_id, ontology,
               object_type):
    """
    Parameters
    ----------
    term_id: str-list
        Identifier to analyze
    ontology: str
        Ontology (knowledge resource).
        Should be one of {'go', 'sbo', 'chebi'}.
    object_type: libsbml.AutoProperty
        Type of model entity. For example, libsbml.Reaction
    """
    self.term_id = term_id
    self.ontology = ontology
    self.object_type = object_type
    self.dag = ONT_TO_G[self.ontology]
    self.possible_roots = ONT_TO_ROOT[self.ontology]
    self.term_to_root = None
    self.consistent = self.getConsistency(inp_term=self.term_id)

  def findRoot(self, inp_term):
    """
    Find the appropriate root for
    the given term, 
    as specified by ONT_TO_ROOT.
    If none is found, return None.
    (which is an error)

    Parameters
    ----------
    inp_term: str
        Identifier (string) to check. 

    Returns
    -------
    one_ances: str, or None
        Identifier of the root for the inp_term. 
    """
    if inp_term in self.dag:
      for one_ances in self.possible_roots:
        has_p = nx.has_path(self.dag, source=inp_term, target=one_ances)
        if has_p:
          return one_ances
    return None

  def getConsistency(self, inp_term):
    """
    Tests if the given term
    or the list of the terms
    are consistent.
    inp_term should be either string or
    a list of strings
    (checked at the beginning).

    Parameters
    ----------
    inp_term: str/str-list
        Identifier (or list of identifiers)

    Returns
    -------
    '': bool
    """
    # first, check if the input term is in correct format
    if isinstance(inp_term, str):
      inp_list = [inp_term]
    elif isinstance(inp_term, list):
      # if it is a list, check whether all terms are string
      if sum([isinstance(t, str) for t in inp_term])==len(inp_term):
        inp_list = inp_term
      else:
      	return False
    else:
      return False
    # if the term(s) are in correct format, check parents (roots)
    par_dict = {one_term:self.findRoot(one_term) for one_term in inp_list}
    pars = [par_dict[k] for k in par_dict.keys()]
    par_types = set([cn.DAG_ROOT_MAP[self.ontology][ele] for ele in pars])
    acceptable_par_types = cn.OBJECT_ONT_MAP_FILT[self.object_type]
    if par_types <= acceptable_par_types:
      self.term_to_root = par_dict
      return True
    else:
      return False

  def getOneTermSpecificity(self, one_term):
    """
    Get specificity score for one term.

    Parameters
    ----------
    one_term: str
         Identifier to check specificity.

    Returns
    -------
    spec_score: float
        Specificity score. 
    """
    if not self.consistent or one_term not in self.dag:
      return None
    # Find appropriate root term
    root_term = self.term_to_root[one_term]
    # add 1 to include itself
    num_ancestors = len(nx.ancestors(self.dag, one_term))+1
    num_all_nodes = len(nx.ancestors(self.dag, root_term))+1
    spec_score = abs(np.log(num_ancestors/num_all_nodes) / np.log(1/num_all_nodes))
    return spec_score

  def getSpecificity(self, inp_term):
    """
    Get specificity score of one or more terms,
    using getOneTermSpecificity().
    Specificity is calculated only if
    the term is already consistent.

    Parameters
    ----------
    inp_term: str/str-list
        Identifier(s) to check specificity.

    Returns
    -------
    '': float / None
        Specificity score (averaged within the same ontology).
    """
    if not self.consistent:
      return None
    if isinstance(inp_term, str):
      inp_list = [inp_term]
    else:
      inp_list = inp_term
    if set(inp_list) <= self.term_to_root.keys():
      res = [self.getOneTermSpecificity(val) for val in inp_list]
      return np.mean(res)
    else:
      print("Calculate consistency first!")
      return None





from setuptools import setup, find_packages

INSTALL_REQUIRES = [
    "coverage==5.5",
    "coveralls",
    "networkx",
    "nose",
    "numpy",
    "pandas",
    "pickle5",
    "python-libsbml",
    "pyyaml",
    "requests"
    ]

def doSetup(install_requires):
  setup(
      name='SBMate',
      version='1.1.3',
      author='Woosub Shin, Joseph L. Hellerstein',
      author_email='woosubs@umich.edu',
      packages=find_packages(exclude=['tests', 'notebooks']),
      url='https://github.com/woosubs/SBMate',
      description='Annotation quality metrics calculator (coverage, consistency, specificity).',
      long_description=open('README.md').read(),
      long_description_content_type='text/markdown',
      package_dir={'SBMate': 'SBMate'},
      install_requires=install_requires,
      include_package_data=True,
      classifiers=[
          'Development Status :: 3 - Alpha',
          'Intended Audience :: Developers',      # Define that your audience are developers
          'Topic :: Scientific/Engineering :: Bio-Informatics',
          'License :: OSI Approved :: MIT License',   # Again, pick a license
          'Programming Language :: Python :: 3.8',
          'Programming Language :: Python :: 3.9',
        ],
      )

if __name__ == '__main__':
  doSetup(INSTALL_REQUIRES)
